@@ -67,14 +67,6 @@ typedef struct TkDisplayFocusInfo {
 				 * for a given application. */
 } DisplayFocusInfo;
 
-/*
- * The following magic value is stored in the "send_event" field of FocusIn
- * and FocusOut events that are generated in this file. This allows us to
- * separate "real" events coming from the server from those that we generated.
- */
-
-#define GENERATED_EVENT_MAGIC	((Bool) 0x547321ac)
-
 /*
  * Debugging support...
  */
@@ -116,16 +108,16 @@ Tk_FocusObjCmd(
     ClientData clientData,	/* Main window associated with interpreter. */
     Tcl_Interp *interp,		/* Current interpreter. */
     int objc,			/* Number of arguments. */
-    Tcl_Obj *CONST objv[])	/* Argument objects. */
+    Tcl_Obj *const objv[])	/* Argument objects. */
 {
-    static CONST char *focusOptions[] = {
+    static const char *const focusOptions[] = {
 	"-displayof", "-force", "-lastfor", NULL
     };
-    Tk_Window tkwin = (Tk_Window) clientData;
-    TkWindow *winPtr = (TkWindow *) clientData;
-    TkWindow *newPtr, *focusWinPtr, *topLevelPtr;
+    Tk_Window tkwin = (Tk_Window)clientData;
+    TkWindow *winPtr = (TkWindow *)clientData;
+    TkWindow *newPtr, *topLevelPtr;
     ToplevelFocusInfo *tlFocusPtr;
-    char *windowName;
+    const char *windowName;
     int index;
 
     /*
@@ -133,9 +125,10 @@ Tk_FocusObjCmd(
      */
 
     if (objc == 1) {
-	focusWinPtr = TkGetFocusWin(winPtr);
-	if (focusWinPtr != NULL) {
-	    Tcl_SetResult(interp, focusWinPtr->pathName, TCL_STATIC);
+	Tk_Window focusWin = (Tk_Window) TkGetFocusWin(winPtr);
+
+	if (focusWin != NULL) {
+	    Tcl_SetObjResult(interp, TkNewWindowObj(focusWin));
 	}
 	return TCL_OK;
     }
@@ -160,9 +153,7 @@ Tk_FocusObjCmd(
 	    if (newPtr == NULL) {
 		return TCL_ERROR;
 	    }
-	    if (!(newPtr->flags & TK_ALREADY_DEAD)) {
-		TkSetFocusWin(newPtr, 0);
-	    }
+	    TkSetFocusWin(newPtr, 0);
 	    return TCL_OK;
 	}
     }
@@ -171,8 +162,8 @@ Tk_FocusObjCmd(
      * We have a subcommand to parse and act upon.
      */
 
-    if (Tcl_GetIndexFromObj(interp, objv[1], focusOptions, "option", 0,
-	    &index) != TCL_OK) {
+    if (Tcl_GetIndexFromObjStruct(interp, objv[1], focusOptions,
+	    sizeof(char *), "option", 0, &index) != TCL_OK) {
     	return TCL_ERROR;
     }
     if (objc != 3) {
@@ -188,7 +179,7 @@ Tk_FocusObjCmd(
 	}
 	newPtr = TkGetFocusWin(newPtr);
 	if (newPtr != NULL) {
-	    Tcl_SetResult(interp, newPtr->pathName, TCL_STATIC);
+	    Tcl_SetObjResult(interp, TkNewWindowObj((Tk_Window) newPtr));
 	}
 	break;
     case 1:			/* -force */
@@ -214,19 +205,19 @@ Tk_FocusObjCmd(
 	    return TCL_ERROR;
 	}
 	for (topLevelPtr = newPtr; topLevelPtr != NULL;
-		topLevelPtr = topLevelPtr->parentPtr)  {
+		topLevelPtr = topLevelPtr->parentPtr) {
 	    if (!(topLevelPtr->flags & TK_TOP_HIERARCHY)) {
 		continue;
 	    }
 	    for (tlFocusPtr = newPtr->mainPtr->tlFocusPtr; tlFocusPtr != NULL;
 		    tlFocusPtr = tlFocusPtr->nextPtr) {
 		if (tlFocusPtr->topLevelPtr == topLevelPtr) {
-		    Tcl_SetResult(interp,
-			    tlFocusPtr->focusWinPtr->pathName, TCL_STATIC);
+		    Tcl_SetObjResult(interp, TkNewWindowObj((Tk_Window)
+			    tlFocusPtr->focusWinPtr));
 		    return TCL_OK;
 		}
 	    }
-	    Tcl_SetResult(interp, topLevelPtr->pathName, TCL_STATIC);
+	    Tcl_SetObjResult(interp, TkNewWindowObj((Tk_Window) topLevelPtr));
 	    return TCL_OK;
 	}
 	break;
@@ -288,8 +279,8 @@ TkFocusFilterEvent(
      * pass the event through to Tk bindings.
      */
 
-    if (eventPtr->xfocus.send_event == GENERATED_EVENT_MAGIC) {
-	eventPtr->xfocus.send_event = 0;
+    if ((eventPtr->xfocus.send_event & GENERATED_FOCUS_EVENT_MAGIC) == GENERATED_FOCUS_EVENT_MAGIC) {
+	eventPtr->xfocus.send_event &= ~GENERATED_FOCUS_EVENT_MAGIC;
 	return 1;
     }
 
@@ -391,7 +382,7 @@ TkFocusFilterEvent(
      * tree, then ignore the event.
      */
 
-    if (TkGrabState(winPtr) == TK_GRAB_EXCLUDED)  {
+    if (TkGrabState(winPtr) == TK_GRAB_EXCLUDED) {
 	return retValue;
     }
 
@@ -424,7 +415,7 @@ TkFocusFilterEvent(
 	}
     }
     if (tlFocusPtr == NULL) {
-	tlFocusPtr = (ToplevelFocusInfo *) ckalloc(sizeof(ToplevelFocusInfo));
+	tlFocusPtr = (ToplevelFocusInfo *)ckalloc(sizeof(ToplevelFocusInfo));
 	tlFocusPtr->topLevelPtr = tlFocusPtr->focusWinPtr = winPtr;
 	tlFocusPtr->nextPtr = winPtr->mainPtr->tlFocusPtr;
 	winPtr->mainPtr->tlFocusPtr = tlFocusPtr;
@@ -498,14 +489,14 @@ TkFocusFilterEvent(
     } else if (eventPtr->type == LeaveNotify) {
 	/*
 	 * If the pointer just left a window for which we automatically
-	 * claimed the focus on enter, move the focus back to the root
-	 * window, where it was before we claimed it above. Note:
+	 * claimed the focus on enter, move the focus back to the root window,
+	 * where it was before we claimed it above. Note:
 	 * dispPtr->implicitWinPtr may not be the same as
-	 * displayFocusPtr->focusWinPtr (e.g. because the "focus" command
-	 * was used to redirect the focus after it arrived at
-	 * dispPtr->implicitWinPtr)!! In addition, we generate events
-	 * because the window manager won't give us a FocusOut event when
-	 * we focus on the root.
+	 * displayFocusPtr->focusWinPtr (e.g. because the "focus" command was
+	 * used to redirect the focus after it arrived at
+	 * dispPtr->implicitWinPtr)!! In addition, we generate events because
+	 * the window manager won't give us a FocusOut event when we focus on
+	 * the root.
 	 */
 
 	if ((dispPtr->implicitWinPtr != NULL)
@@ -552,12 +543,25 @@ TkSetFocusWin(
     TkWindow *topLevelPtr;
     int allMapped, serial;
 
+    /*
+     * Don't set focus if window is already dead. [Bug 3574708]
+     */
+
+    if (winPtr->flags & TK_ALREADY_DEAD) {
+	return;
+    }
+
+    /*
+     * Get the current focus window with the same display and application
+     * as winPtr.
+     */
+
     displayFocusPtr = FindDisplayFocusInfo(winPtr->mainPtr, winPtr->dispPtr);
 
     /*
-     * If force is set, we should make sure we grab the focus regardless of
-     * the current focus window since under Windows, we may need to take
-     * control away from another application.
+     * Do nothing if the window already has focus and force is not set.  If
+     * force is set, we need to grab the focus, since under Windows or macOS
+     * this may involve taking control away from another application.
      */
 
     if (winPtr == displayFocusPtr->focusWinPtr && !force) {
@@ -565,14 +569,15 @@ TkSetFocusWin(
     }
 
     /*
-     * Find the top-level window for winPtr, then find (or create) a record
-     * for the top-level. Also see whether winPtr and all its ancestors are
+     * Find the toplevel window for winPtr, then find (or create) a record
+     * for the toplevel. Also see whether winPtr and all its ancestors are
      * mapped.
      */
 
     allMapped = 1;
-    for (topLevelPtr = winPtr; ; topLevelPtr = topLevelPtr->parentPtr)  {
+    for (topLevelPtr = winPtr; ; topLevelPtr = topLevelPtr->parentPtr) {
 	if (topLevelPtr == NULL) {
+
 	    /*
 	     * The window is being deleted. No point in worrying about giving
 	     * it the focus.
@@ -589,24 +594,22 @@ TkSetFocusWin(
     }
 
     /*
-     * If the new focus window isn't mapped, then we can't focus on it (X will
-     * generate an error, for example). Instead, create an event handler that
-     * will set the focus to this window once it gets mapped. At the same
-     * time, delete any old handler that might be around; it's no longer
-     * relevant.
+     * If any ancestor of the new focus window isn't mapped, then we can't set
+     * focus for it (X will generate an error, for example). Instead, create
+     * an event handler that will set the focus to this window once it gets
+     * mapped. At the same time, delete any old handler that might be around;
+     * it's no longer relevant.
      */
 
     if (displayFocusPtr->focusOnMapPtr != NULL) {
-	Tk_DeleteEventHandler(
-		(Tk_Window) displayFocusPtr->focusOnMapPtr,
-		StructureNotifyMask, FocusMapProc,
-		(ClientData) displayFocusPtr->focusOnMapPtr);
+	Tk_DeleteEventHandler((Tk_Window) displayFocusPtr->focusOnMapPtr,
+		VisibilityChangeMask, FocusMapProc,
+		displayFocusPtr->focusOnMapPtr);
 	displayFocusPtr->focusOnMapPtr = NULL;
     }
     if (!allMapped) {
-	Tk_CreateEventHandler((Tk_Window) winPtr,
-		VisibilityChangeMask, FocusMapProc,
-		(ClientData) winPtr);
+	Tk_CreateEventHandler((Tk_Window) winPtr, VisibilityChangeMask,
+		FocusMapProc, winPtr);
 	displayFocusPtr->focusOnMapPtr = winPtr;
 	displayFocusPtr->forceFocus = force;
 	return;
@@ -619,35 +622,44 @@ TkSetFocusWin(
 	}
     }
     if (tlFocusPtr == NULL) {
-	tlFocusPtr = (ToplevelFocusInfo *) ckalloc(sizeof(ToplevelFocusInfo));
+	tlFocusPtr = (ToplevelFocusInfo *)ckalloc(sizeof(ToplevelFocusInfo));
 	tlFocusPtr->topLevelPtr = topLevelPtr;
 	tlFocusPtr->nextPtr = winPtr->mainPtr->tlFocusPtr;
 	winPtr->mainPtr->tlFocusPtr = tlFocusPtr;
     }
     tlFocusPtr->focusWinPtr = winPtr;
 
-    /*
-     * Reset the window system's focus window and generate focus events, with
-     * two special cases:
-     *
-     * 1. If the application is embedded and doesn't currently have the focus,
-     *    don't set the focus directly. Instead, see if the embedding code can
-     *    claim the focus from the enclosing container.
-     * 2. Otherwise, if the application doesn't currently have the focus,
-     *    don't change the window system's focus unless it was already in this
-     *    application or "force" was specified.
-     */
+    if (topLevelPtr->flags & TK_EMBEDDED &&
+        (displayFocusPtr->focusWinPtr == NULL)) {
+
+	/*
+	 * We are assigning focus to an embedded toplevel.  The platform
+	 * specific function TkpClaimFocus needs to handle the job of
+	 * assigning focus to the container, since we have no way to find the
+	 * contaiuner.
+	 */
 
-    if ((topLevelPtr->flags & TK_EMBEDDED)
-	    && (displayFocusPtr->focusWinPtr == NULL)) {
 	TkpClaimFocus(topLevelPtr, force);
     } else if ((displayFocusPtr->focusWinPtr != NULL) || force) {
+
 	/*
-	 * Generate events to shift focus between Tk windows. We do this
-	 * regardless of what TkpChangeFocus does with the real X focus so
-	 * that Tk widgets track focus commands when there is no window
-	 * manager. GenerateFocusEvents will set up a serial number marker so
-	 * we discard focus events that are triggered by the ChangeFocus.
+	 * If we are forcing removal of focus from a container hosting a
+	 * toplevel from a different application, clear the focus in that
+	 * application.
+	 */
+
+    	if (force) {
+	    TkWindow *focusPtr = winPtr->dispPtr->focusPtr;
+	    if (focusPtr && focusPtr->mainPtr != winPtr->mainPtr) {
+		DisplayFocusInfo *displayFocusPtr2 = FindDisplayFocusInfo(
+		    focusPtr->mainPtr, focusPtr->dispPtr);
+		displayFocusPtr2->focusWinPtr = NULL;
+	    }
+    	}
+
+	/*
+	 * Call the platform specific function TkpChangeFocus to move the
+	 * window manager's focus to a new toplevel.
 	 */
 
 	serial = TkpChangeFocus(TkpGetWrapperWindow(topLevelPtr), force);
@@ -795,7 +807,7 @@ TkFocusKeyEvent(
 
 void
 TkFocusDeadWindow(
-    register TkWindow *winPtr)	/* Information about the window that is being
+    TkWindow *winPtr)	/* Information about the window that is being
 				 * deleted. */
 {
     ToplevelFocusInfo *tlFocusPtr, *prevPtr;
@@ -843,7 +855,7 @@ TkFocusDeadWindow(
 	    } else {
 		prevPtr->nextPtr = tlFocusPtr->nextPtr;
 	    }
-	    ckfree((char *) tlFocusPtr);
+	    ckfree(tlFocusPtr);
 	    break;
 	} else if (winPtr == tlFocusPtr->focusWinPtr) {
 	    /*
@@ -916,7 +928,7 @@ GenerateFocusEvents(
     }
 
     event.xfocus.serial = LastKnownRequestProcessed(winPtr->display);
-    event.xfocus.send_event = GENERATED_EVENT_MAGIC;
+    event.xfocus.send_event = GENERATED_FOCUS_EVENT_MAGIC;
     event.xfocus.display = winPtr->display;
     event.xfocus.mode = NotifyNormal;
     TkInOutEvents(&event, sourcePtr, destPtr, FocusOut, FocusIn,
@@ -949,7 +961,7 @@ FocusMapProc(
     ClientData clientData,	/* Toplevel window. */
     XEvent *eventPtr)		/* Information about event. */
 {
-    TkWindow *winPtr = (TkWindow *) clientData;
+    TkWindow *winPtr = (TkWindow *)clientData;
     DisplayFocusInfo *displayFocusPtr;
 
     if (eventPtr->type == VisibilityNotify) {
@@ -1003,7 +1015,7 @@ FindDisplayFocusInfo(
      * The record doesn't exist yet. Make a new one.
      */
 
-    displayFocusPtr = (DisplayFocusInfo *) ckalloc(sizeof(DisplayFocusInfo));
+    displayFocusPtr = (DisplayFocusInfo *)ckalloc(sizeof(DisplayFocusInfo));
     displayFocusPtr->dispPtr = dispPtr;
     displayFocusPtr->focusWinPtr = NULL;
     displayFocusPtr->focusOnMapPtr = NULL;
@@ -1039,13 +1051,13 @@ TkFocusFree(
 	DisplayFocusInfo *displayFocusPtr = mainPtr->displayFocusPtr;
 
 	mainPtr->displayFocusPtr = mainPtr->displayFocusPtr->nextPtr;
-	ckfree((char *) displayFocusPtr);
+	ckfree(displayFocusPtr);
     }
     while (mainPtr->tlFocusPtr != NULL) {
 	ToplevelFocusInfo *tlFocusPtr = mainPtr->tlFocusPtr;
 
 	mainPtr->tlFocusPtr = mainPtr->tlFocusPtr->nextPtr;
-	ckfree((char *) tlFocusPtr);
+	ckfree(tlFocusPtr);
     }
 }
 
@@ -1054,8 +1066,8 @@ TkFocusFree(
  *
  * TkFocusSplit --
  *
- *	Adjust focus window for a newly managed toplevel, thus splitting
- *      the toplevel into two toplevels.
+ *	Adjust focus window for a newly managed toplevel, thus splitting the
+ *	toplevel into two toplevels.
  *
  * Results:
  *	None.
@@ -1067,29 +1079,29 @@ TkFocusFree(
  */
 
 void
-TkFocusSplit(winPtr)
-    TkWindow *winPtr;          /* Window is the new toplevel
-                                * Any focus point at or below window
-				* must be moved to this new toplevel */
+TkFocusSplit(
+    TkWindow *winPtr)		/* Window is the new toplevel. Any focus point
+				 * at or below window must be moved to this
+				 * new toplevel. */
 {
     ToplevelFocusInfo *tlFocusPtr;
-    TkWindow *topLevelPtr;
-    TkWindow *subWinPtr;
+    TkWindow *topLevelPtr, *subWinPtr;
 
     FindDisplayFocusInfo(winPtr->mainPtr, winPtr->dispPtr);
 
     /*
-     * Find the top-level window for winPtr, then find (or create)
-     * a record for the top-level.  Also see whether winPtr and all its
-     * ancestors are mapped.
+     * Find the top-level window for winPtr, then find (or create) a record
+     * for the top-level. Also see whether winPtr and all its ancestors are
+     * mapped.
      */
 
-    for (topLevelPtr = winPtr; ; topLevelPtr = topLevelPtr->parentPtr)  {
+    for (topLevelPtr = winPtr; ; topLevelPtr = topLevelPtr->parentPtr) {
 	if (topLevelPtr == NULL) {
 	    /*
-	     * The window is being deleted.  No point in worrying about
-	     * giving it the focus.
+	     * The window is being deleted. No point in worrying about giving
+	     * it the focus.
 	     */
+
 	    return;
 	}
 	if (topLevelPtr->flags & TK_TOP_HIERARCHY) {
@@ -1097,37 +1109,57 @@ TkFocusSplit(winPtr)
 	}
     }
 
-    /* Search all focus records to find child windows of winPtr */
+    /*
+     * Search all focus records to find child windows of winPtr.
+     */
+
     for (tlFocusPtr = winPtr->mainPtr->tlFocusPtr; tlFocusPtr != NULL;
-	 tlFocusPtr = tlFocusPtr->nextPtr) {
+	    tlFocusPtr = tlFocusPtr->nextPtr) {
 	if (tlFocusPtr->topLevelPtr == topLevelPtr) {
 	    break;
 	}
     }
 
     if (tlFocusPtr == NULL) {
-	/* No focus record for this toplevel, nothing to do. */
+	/*
+	 * No focus record for this toplevel, nothing to do.
+	 */
+
 	return;
     }
 
-    /* See if current focusWin is child of the new toplevel */
-    for (subWinPtr = tlFocusPtr->focusWinPtr; 
-	 subWinPtr && subWinPtr != winPtr && subWinPtr != topLevelPtr;
-	 subWinPtr = subWinPtr->parentPtr) {}
+    /*
+     * See if current focusWin is child of the new toplevel.
+     */
+
+    for (subWinPtr = tlFocusPtr->focusWinPtr;
+	    subWinPtr && subWinPtr != winPtr && subWinPtr != topLevelPtr;
+	    subWinPtr = subWinPtr->parentPtr) {
+	/* EMPTY */
+    }
 
     if (subWinPtr == winPtr) {
-	/* Move focus to new toplevel */
-	ToplevelFocusInfo *newTlFocusPtr;
+	/*
+	 * Move focus to new toplevel.
+	 */
+
+	ToplevelFocusInfo *newTlFocusPtr = (ToplevelFocusInfo *)ckalloc(sizeof(ToplevelFocusInfo));
 
-	newTlFocusPtr = (ToplevelFocusInfo *) ckalloc(sizeof(ToplevelFocusInfo));
 	newTlFocusPtr->topLevelPtr = winPtr;
 	newTlFocusPtr->focusWinPtr = tlFocusPtr->focusWinPtr;
 	newTlFocusPtr->nextPtr = winPtr->mainPtr->tlFocusPtr;
 	winPtr->mainPtr->tlFocusPtr = newTlFocusPtr;
-	/* Move old toplevel's focus to the toplevel itself */
+
+	/*
+	 * Move old toplevel's focus to the toplevel itself.
+	 */
+
 	tlFocusPtr->focusWinPtr = topLevelPtr;
     }
-    /* If it's not, then let focus progress naturally */
+
+    /*
+     * If it's not, then let focus progress naturally.
+     */
 }
 
 /*
@@ -1147,28 +1179,28 @@ TkFocusSplit(winPtr)
  */
 
 void
-TkFocusJoin(winPtr)
-    TkWindow *winPtr;          /* Window is no longer a toplevel */
+TkFocusJoin(
+    TkWindow *winPtr)		/* Window is no longer a toplevel. */
 {
-    ToplevelFocusInfo *tlFocusPtr;
-    ToplevelFocusInfo *tmpPtr;
+    ToplevelFocusInfo *tlFocusPtr, *tmpPtr;
 
     /*
      * Remove old toplevel record
      */
+
     if (winPtr && winPtr->mainPtr && winPtr->mainPtr->tlFocusPtr
-	&& winPtr->mainPtr->tlFocusPtr->topLevelPtr == winPtr) {
+	    && winPtr->mainPtr->tlFocusPtr->topLevelPtr == winPtr) {
 	tmpPtr = winPtr->mainPtr->tlFocusPtr;
 	winPtr->mainPtr->tlFocusPtr = tmpPtr->nextPtr;
-	ckfree((char *)tmpPtr);
-    } else {
+	ckfree(tmpPtr);
+    } else if (winPtr && winPtr->mainPtr) {
 	for (tlFocusPtr = winPtr->mainPtr->tlFocusPtr; tlFocusPtr != NULL;
-	     tlFocusPtr = tlFocusPtr->nextPtr) {
+		tlFocusPtr = tlFocusPtr->nextPtr) {
 	    if (tlFocusPtr->nextPtr &&
-		tlFocusPtr->nextPtr->topLevelPtr == winPtr) {
+		    tlFocusPtr->nextPtr->topLevelPtr == winPtr) {
 		tmpPtr = tlFocusPtr->nextPtr;
 		tlFocusPtr->nextPtr = tmpPtr->nextPtr;
-		ckfree((char *)tmpPtr);
+		ckfree(tmpPtr);
 		break;
 	    }
 	}
